@@ -1,5 +1,5 @@
 /* snac - A simple, minimalistic ActivityPub instance */
-/* copyright (c) 2022 - 2023 grunfink et al. / MIT license */
+/* copyright (c) 2022 - 2024 grunfink et al. / MIT license */
 
 #include "xs.h"
 #include "xs_io.h"
@@ -9,6 +9,8 @@
 #include "xs_mime.h"
 #include "xs_time.h"
 #include "xs_openssl.h"
+#include "xs_fcgi.h"
+#include "xs_html.h"
 
 #include "snac.h"
 
@@ -20,11 +22,39 @@
 
 #include <sys/resource.h> // for getrlimit()
 
+#include <sys/mman.h>
+
 #ifdef USE_POLL_FOR_SLEEP
 #include <poll.h>
 #endif
 
-int srv_running = 0;
+/** server state **/
+srv_state *p_state = NULL;
+
+
+/** job control **/
+
+/* mutex to access the lists of jobs */
+static pthread_mutex_t job_mutex;
+
+/* semaphore to trigger job processing */
+static sem_t *job_sem;
+
+typedef struct job_fifo_item {
+    struct job_fifo_item *next;
+    xs_val *job;
+} job_fifo_item;
+
+static job_fifo_item *job_fifo_first = NULL;
+static job_fifo_item *job_fifo_last  = NULL;
+
+
+/** other global data **/
+
+static jmp_buf on_break;
+
+
+/** code **/
 
 /* nodeinfo 2.0 template */
 const char *nodeinfo_2_0_template = ""
@@ -44,14 +74,14 @@ xs_str *nodeinfo_2_0(void)
     int n_uhyear = 0;
     int n_posts  = 0;
     xs *users = user_list();
-    xs_list *p;
+    xs_list *p = users;
     char *v;
+    double now = (double)time(NULL);
 
-    p = users;
     while (xs_list_iter(&p, &v)) {
         /* build the full path name to the last usage log */
         xs *llfn = xs_fmt("%s/user/%s/lastlog.txt", srv_basedir, v);
-        double llsecs = (double)time(NULL) - mtime(llfn);
+        double llsecs = now - mtime(llfn);
 
         if (llsecs < 60 * 60 * 24 * 30 * 6) {
             n_uhyear++;
@@ -63,7 +93,7 @@ xs_str *nodeinfo_2_0(void)
         n_utotal++;
 
         /* build the file to each user public.idx */
-        xs *pidxfn = xs_fmt("%s/user/%s/private.idx", srv_basedir, v);
+        xs *pidxfn = xs_fmt("%s/user/%s/public.idx", srv_basedir, v);
         n_posts += index_len(pidxfn);
     }
 
@@ -95,32 +125,37 @@ static xs_str *greeting_html(void)
 
         /* does it have a %userlist% mark? */
         if (xs_str_in(s, "%userlist%") != -1) {
-            const char *host = xs_dict_get(srv_config, "host");
+            char *host = xs_dict_get(srv_config, "host");
             xs *list = user_list();
-            xs_list *p;
+            xs_list *p = list;
             xs_str *uid;
-            xs *ul = xs_str_new("<ul class=\"snac-user-list\">\n");
+
+            xs_html *ul = xs_html_tag("ul",
+                xs_html_attr("class", "snac-user-list"));
 
             p = list;
             while (xs_list_iter(&p, &uid)) {
                 snac user;
 
                 if (user_open(&user, uid)) {
-                    xs *uname = encode_html(xs_dict_get(user.config, "name"));
-
-                    xs *u = xs_fmt(
-                        "<li><a href=\"%s\">@%s@%s (%s)</a></li>\n",
-                            user.actor, uid, host, uname);
-
-                    ul = xs_str_cat(ul, u);
+                    xs_html_add(ul,
+                        xs_html_tag("li",
+                            xs_html_tag("a",
+                                xs_html_attr("href", user.actor),
+                                xs_html_text("@"),
+                                xs_html_text(uid),
+                                xs_html_text("@"),
+                                xs_html_text(host),
+                                xs_html_text(" ("),
+                                xs_html_text(xs_dict_get(user.config, "name")),
+                                xs_html_text(")"))));
 
                     user_free(&user);
                 }
             }
 
-            ul = xs_str_cat(ul, "</ul>\n");
-
-            s = xs_replace_i(s, "%userlist%", ul);
+            xs *s1 = xs_html_render(ul);
+            s = xs_replace_i(s, "%userlist%", s1);
         }
     }
 
@@ -138,9 +173,33 @@ int server_get_handler(xs_dict *req, const char *q_path,
 
     /* is it the server root? */
     if (*q_path == '\0') {
+        xs_dict *q_vars = xs_dict_get(req, "q_vars");
+        char *t = NULL;
+
+        if (xs_type(q_vars) == XSTYPE_DICT && (t = xs_dict_get(q_vars, "t"))) {
+            int skip = 0;
+            int show = xs_number_get(xs_dict_get(srv_config, "max_timeline_entries"));
+            char *v;
+
+            if ((v = xs_dict_get(q_vars, "skip")) != NULL)
+                skip = atoi(v);
+            if ((v = xs_dict_get(q_vars, "show")) != NULL)
+                show = atoi(v);
+
+            xs *tl = tag_search(t, skip, show + 1);
+            int more = 0;
+            if (xs_list_len(tl) >= show + 1) {
+                /* drop the last one */
+                tl = xs_list_del(tl, -1);
+                more = 1;
+            }
+
+            *body = html_timeline(NULL, tl, 0, skip, show, more, t);
+        }
+        else
         if (xs_type(xs_dict_get(srv_config, "show_instance_timeline")) == XSTYPE_TRUE) {
             xs *tl = timeline_instance_list(0, 30);
-            *body = html_timeline(NULL, tl, 0, 0, 0, 0);
+            *body = html_timeline(NULL, tl, 0, 0, 0, 0, NULL);
         }
         else
             *body = greeting_html();
@@ -199,8 +258,12 @@ void httpd_connection(FILE *f)
     xs *etag     = NULL;
     int p_size   = 0;
     char *p;
+    int fcgi_id;
 
-    req = xs_httpd_request(f, &payload, &p_size);
+    if (p_state->use_fcgi)
+        req = xs_fcgi_request(f, &payload, &p_size, &fcgi_id);
+    else
+        req = xs_httpd_request(f, &payload, &p_size);
 
     if (req == NULL) {
         /* probably because a timeout */
@@ -208,8 +271,13 @@ void httpd_connection(FILE *f)
         return;
     }
 
-    method = xs_dict_get(req, "method");
-    q_path = xs_dup(xs_dict_get(req, "path"));
+    if (!(method = xs_dict_get(req, "method")) || !(p = xs_dict_get(req, "path"))) {
+        /* missing needed headers; discard */
+        fclose(f);
+        return;
+    }
+
+    q_path = xs_dup(p);
 
     /* crop the q_path from leading / and the prefix */
     if (xs_endswith(q_path, "/"))
@@ -276,6 +344,14 @@ void httpd_connection(FILE *f)
     if (strcmp(method, "OPTIONS") == 0) {
         status = 200;
     }
+    else
+    if (strcmp(method, "DELETE") == 0) {
+#ifndef NO_MASTODON_API
+        if (status == 0)
+            status = mastoapi_delete_handler(req, q_path,
+                    &body, &b_size, &ctype);
+#endif
+    }
 
     /* unattended? it's an error */
     if (status == 0) {
@@ -330,14 +406,17 @@ void httpd_connection(FILE *f)
     headers = xs_dict_append(headers, "access-control-allow-origin", "*");
     headers = xs_dict_append(headers, "access-control-allow-headers", "*");
 
-    xs_httpd_response(f, status, headers, body, b_size);
+    if (p_state->use_fcgi)
+        xs_fcgi_response(f, status, headers, body, b_size, fcgi_id);
+    else
+        xs_httpd_response(f, status, headers, body, b_size);
 
     fclose(f);
 
     srv_archive("RECV", NULL, req, payload, p_size, status, headers, body, b_size);
 
     /* JSON validation check */
-    if (strcmp(ctype, "application/json") == 0) {
+    if (!xs_is_null(body) && strcmp(ctype, "application/json") == 0) {
         xs *j = xs_json_loads(body);
 
         if (j == NULL) {
@@ -350,36 +429,6 @@ void httpd_connection(FILE *f)
 }
 
 
-static jmp_buf on_break;
-
-
-void term_handler(int s)
-{
-    (void)s;
-
-    longjmp(on_break, 1);
-}
-
-
-/** job control **/
-
-/* mutex to access the lists of jobs */
-static pthread_mutex_t job_mutex;
-
-/* semaphre to trigger job processing */
-static sem_t *job_sem;
-
-/* fifo of jobs */
-xs_list *job_fifo = NULL;
-
-
-int job_fifo_ready(void)
-/* returns true if the job fifo is ready */
-{
-    return job_fifo != NULL;
-}
-
-
 void job_post(const xs_val *job, int urgent)
 /* posts a job for the threads to process it */
 {
@@ -387,20 +436,34 @@ void job_post(const xs_val *job, int urgent)
         /* lock the mutex */
         pthread_mutex_lock(&job_mutex);
 
-        /* add to the fifo */
-        if (job_fifo != NULL) {
-            if (urgent)
-                job_fifo = xs_list_insert(job_fifo, 0, job);
-            else
-                job_fifo = xs_list_append(job_fifo, job);
+        job_fifo_item *i = xs_realloc(NULL, sizeof(job_fifo_item));
+        *i = (job_fifo_item){ NULL, xs_dup(job) };
+
+        if (job_fifo_first == NULL)
+            job_fifo_first = job_fifo_last = i;
+        else
+        if (urgent) {
+            /* prepend */
+            i->next = job_fifo_first;
+            job_fifo_first = i;
         }
+        else {
+            /* append */
+            job_fifo_last->next = i;
+            job_fifo_last = i;
+        }
+
+        p_state->job_fifo_size++;
+
+        if (p_state->job_fifo_size > p_state->peak_job_fifo_size)
+            p_state->peak_job_fifo_size = p_state->job_fifo_size;
 
         /* unlock the mutex */
         pthread_mutex_unlock(&job_mutex);
-    }
 
-    /* ask for someone to attend it */
-    sem_post(job_sem);
+        /* ask for someone to attend it */
+        sem_post(job_sem);
+    }
 }
 
 
@@ -414,8 +477,19 @@ void job_wait(xs_val **job)
         pthread_mutex_lock(&job_mutex);
 
         /* dequeue */
-        if (job_fifo != NULL)
-            job_fifo = xs_list_shift(job_fifo, job);
+        job_fifo_item *i = job_fifo_first;
+
+        if (i != NULL) {
+            job_fifo_first = i->next;
+
+            if (job_fifo_first == NULL)
+                job_fifo_last = NULL;
+
+            *job = i->job;
+            xs_free(i);
+
+            p_state->job_fifo_size--;
+        }
 
         /* unlock the mutex */
         pthread_mutex_unlock(&job_mutex);
@@ -423,10 +497,6 @@ void job_wait(xs_val **job)
 }
 
 
-#ifndef MAX_THREADS
-#define MAX_THREADS 256
-#endif
-
 static void *job_thread(void *arg)
 /* job thread */
 {
@@ -437,17 +507,22 @@ static void *job_thread(void *arg)
     for (;;) {
         xs *job = NULL;
 
+        p_state->th_state[pid] = THST_WAIT;
+
         job_wait(&job);
 
-        srv_debug(2, xs_fmt("job thread %d wake up", pid));
+        if (job == NULL) /* corrupted message? */
+            continue;
 
-        if (job == NULL)
+        if (xs_type(job) == XSTYPE_FALSE) /* special message: exit */
             break;
-
+        else
         if (xs_type(job) == XSTYPE_DATA) {
             /* it's a socket */
             FILE *f = NULL;
 
+            p_state->th_state[pid] = THST_IN;
+
             xs_data_get(&f, job);
 
             if (f != NULL)
@@ -455,10 +530,14 @@ static void *job_thread(void *arg)
         }
         else {
             /* it's a q_item */
+            p_state->th_state[pid] = THST_QUEUE;
+
             process_queue_item(job);
         }
     }
 
+    p_state->th_state[pid] = THST_STOP;
+
     srv_debug(1, xs_fmt("job thread %d stopped", pid));
 
     return NULL;
@@ -480,10 +559,12 @@ static void *background_thread(void *arg)
 
     srv_log(xs_fmt("background thread started"));
 
-    while (srv_running) {
+    while (p_state->srv_running) {
         time_t t;
         int cnt = 0;
 
+        p_state->th_state[0] = THST_QUEUE;
+
         {
             xs *list = user_list();
             char *p, *uid;
@@ -516,6 +597,8 @@ static void *background_thread(void *arg)
         if (cnt == 0) {
             /* sleep 3 seconds */
 
+            p_state->th_state[0] = THST_WAIT;
+
 #ifdef USE_POLL_FOR_SLEEP
             poll(NULL, 0, 3 * 1000);
 #else
@@ -531,12 +614,90 @@ static void *background_thread(void *arg)
         }
     }
 
+    p_state->th_state[0] = THST_STOP;
+
     srv_log(xs_fmt("background thread stopped"));
 
     return NULL;
 }
 
 
+void term_handler(int s)
+{
+    (void)s;
+
+    longjmp(on_break, 1);
+}
+
+
+srv_state *srv_state_op(xs_str **fname, int op)
+/* opens or deletes the shared memory object */
+{
+    int fd;
+    srv_state *ss = NULL;
+
+    if (*fname == NULL)
+        *fname = xs_fmt("/%s_snac_state", xs_dict_get(srv_config, "host"));
+
+    switch (op) {
+    case 0: /* open for writing */
+        if ((fd = shm_open(*fname, O_CREAT | O_RDWR, 0666)) != -1) {
+            ftruncate(fd, sizeof(*ss));
+
+            if ((ss = mmap(0, sizeof(*ss), PROT_READ | PROT_WRITE,
+                 MAP_SHARED, fd, 0)) == MAP_FAILED)
+                ss = NULL;
+
+            close(fd);
+        }
+
+        if (ss == NULL) {
+            /* shared memory error: just create a plain structure */
+            srv_log(xs_fmt("warning: shm object error (%s)", strerror(errno)));
+            ss = malloc(sizeof(*ss));
+        }
+
+        /* init structure */
+        *ss = (srv_state){0};
+        ss->s_size = sizeof(*ss);
+
+        break;
+
+    case 1: /* open for reading */
+        if ((fd = shm_open(*fname, O_RDONLY, 0666)) != -1) {
+            if ((ss = mmap(0, sizeof(*ss), PROT_READ, MAP_SHARED, fd, 0)) == MAP_FAILED)
+                ss = NULL;
+
+            close(fd);
+        }
+
+        if (ss == NULL) {
+            /* shared memory error */
+            srv_log(xs_fmt("error: shm object error (%s) server not running?", strerror(errno)));
+        }
+        else
+        if (ss->s_size != sizeof(*ss)) {
+            srv_log(xs_fmt("error: struct size mismatch (%d != %d)",
+                ss->s_size, sizeof(*ss)));
+
+            munmap(ss, sizeof(*ss));
+
+            ss = NULL;
+        }
+
+        break;
+
+    case 2: /* unlink */
+        if (*fname)
+            shm_unlink(*fname);
+
+        break;
+    }
+
+    return ss;
+}
+
+
 void httpd(void)
 /* starts the server */
 {
@@ -544,12 +705,18 @@ void httpd(void)
     const char *port;
     int rs;
     pthread_t threads[MAX_THREADS] = {0};
-    int n_threads = 0;
     int n;
-    time_t start_time = time(NULL);
-    char sem_name[24];
+    xs *sem_name = NULL;
+    xs *shm_name = NULL;
     sem_t anon_job_sem;
 
+    /* setup the server stat structure */
+    p_state = srv_state_op(&shm_name, 0);
+
+    p_state->srv_start_time = time(NULL);
+
+    p_state->use_fcgi = xs_type(xs_dict_get(srv_config, "fastcgi")) == XSTYPE_TRUE;
+
     address = xs_dict_get(srv_config, "address");
     port    = xs_number_str(xs_dict_get(srv_config, "port"));
 
@@ -558,13 +725,14 @@ void httpd(void)
         return;
     }
 
-    srv_running = 1;
+    p_state->srv_running = 1;
 
     signal(SIGPIPE, SIG_IGN);
     signal(SIGTERM, term_handler);
     signal(SIGINT,  term_handler);
 
-    srv_log(xs_fmt("httpd start %s:%s %s", address, port, USER_AGENT));
+    srv_log(xs_fmt("httpd%s start %s:%s %s", p_state->use_fcgi ? " (FastCGI)" : "",
+                    address, port, USER_AGENT));
 
     /* show the number of usable file descriptors */
     struct rlimit r;
@@ -574,7 +742,7 @@ void httpd(void)
 
     /* initialize the job control engine */
     pthread_mutex_init(&job_mutex, NULL);
-    snprintf(sem_name, sizeof(sem_name), "/job_%d", getpid());
+    sem_name = xs_fmt("/job_%d", getpid());
     job_sem = sem_open(sem_name, O_CREAT, 0644, 0);
 
     if (job_sem == NULL) {
@@ -588,35 +756,33 @@ void httpd(void)
         return;
     }
 
-    job_fifo = xs_list_new();
-
     /* initialize sleep control */
     pthread_mutex_init(&sleep_mutex, NULL);
     pthread_cond_init(&sleep_cond, NULL);
 
-    n_threads = xs_number_get(xs_dict_get(srv_config, "num_threads"));
+    p_state->n_threads = xs_number_get(xs_dict_get(srv_config, "num_threads"));
 
 #ifdef _SC_NPROCESSORS_ONLN
-    if (n_threads == 0) {
+    if (p_state->n_threads == 0) {
         /* get number of CPUs on the machine */
-        n_threads = sysconf(_SC_NPROCESSORS_ONLN);
+        p_state->n_threads = sysconf(_SC_NPROCESSORS_ONLN);
     }
 #endif
 
-    if (n_threads < 4)
-        n_threads = 4;
+    if (p_state->n_threads < 4)
+        p_state->n_threads = 4;
 
-    if (n_threads > MAX_THREADS)
-        n_threads = MAX_THREADS;
+    if (p_state->n_threads > MAX_THREADS)
+        p_state->n_threads = MAX_THREADS;
 
-    srv_debug(0, xs_fmt("using %d threads", n_threads));
+    srv_debug(0, xs_fmt("using %d threads", p_state->n_threads));
 
     /* thread #0 is the background thread */
     pthread_create(&threads[0], NULL, background_thread, NULL);
 
     /* the rest of threads are for job processing */
     char *ptr = (char *) 0x1;
-    for (n = 1; n < n_threads; n++)
+    for (n = 1; n < p_state->n_threads; n++)
         pthread_create(&threads[n], NULL, job_thread, ptr++);
 
     if (setjmp(on_break) == 0) {
@@ -632,24 +798,24 @@ void httpd(void)
         }
     }
 
-    srv_running = 0;
+    p_state->srv_running = 0;
 
-    /* send as many empty jobs as working threads */
-    for (n = 1; n < n_threads; n++)
-        job_post(NULL, 0);
+    /* send as many exit jobs as working threads */
+    for (n = 1; n < p_state->n_threads; n++)
+        job_post(xs_stock_false, 0);
 
     /* wait for all the threads to exit */
-    for (n = 0; n < n_threads; n++)
+    for (n = 0; n < p_state->n_threads; n++)
         pthread_join(threads[n], NULL);
 
-    pthread_mutex_lock(&job_mutex);
-    job_fifo = xs_free(job_fifo);
-    pthread_mutex_unlock(&job_mutex);
-
     sem_close(job_sem);
     sem_unlink(sem_name);
 
-    xs *uptime = xs_str_time_diff(time(NULL) - start_time);
+    srv_state_op(&shm_name, 2);
 
-    srv_log(xs_fmt("httpd stop %s:%s (run time: %s)", address, port, uptime));
+    xs *uptime = xs_str_time_diff(time(NULL) - p_state->srv_start_time);
+
+    srv_log(xs_fmt("httpd%s stop %s:%s (run time: %s)",
+                p_state->use_fcgi ? " (FastCGI)" : "",
+                address, port, uptime));
 }
