@@ -1,9 +1,10 @@
 /* snac - A simple, minimalistic ActivityPub instance */
-/* copyright (c) 2022 - 2023 grunfink et al. / MIT license */
+/* copyright (c) 2022 - 2024 grunfink et al. / MIT license */
 
 #ifndef NO_MASTODON_API
 
 #include "xs.h"
+#include "xs_hex.h"
 #include "xs_openssl.h"
 #include "xs_json.h"
 #include "xs_io.h"
@@ -278,8 +279,14 @@ int oauth_post_handler(const xs_dict *req, const char *q_path,
                     xs *code = random_str();
 
                     xs_free(*body);
-                    *body = xs_fmt("%s?code=%s", redir, code);
-                    status = 303;
+
+                    if (strcmp(redir, "urn:ietf:wg:oauth:2.0:oob") == 0) {
+                        *body = xs_dup(code);
+                    }
+                    else {
+                        *body = xs_fmt("%s?code=%s", redir, code);
+                        status = 303;
+                    }
 
                     /* if there is a state, add it */
                     if (!xs_is_null(state) && *state) {
@@ -487,16 +494,20 @@ xs_str *mastoapi_id(const xs_dict *msg)
 xs_dict *mastoapi_account(const xs_dict *actor)
 /* converts an ActivityPub actor to a Mastodon account */
 {
-    xs_dict *acct     = xs_dict_new();
+    const char *id  = xs_dict_get(actor, "id");
+    const char *pub = xs_dict_get(actor, "published");
+
+    if (xs_type(id) != XSTYPE_STRING)
+        return NULL;
+
     const char *prefu = xs_dict_get(actor, "preferredUsername");
 
     const char *display_name = xs_dict_get(actor, "name");
     if (xs_is_null(display_name) || *display_name == '\0')
         display_name = prefu;
 
-    const char *id  = xs_dict_get(actor, "id");
-    const char *pub = xs_dict_get(actor, "published");
-    xs *acct_md5 = xs_md5_hex(id, strlen(id));
+    xs_dict *acct = xs_dict_new();
+    xs *acct_md5  = xs_md5_hex(id, strlen(id));
     acct = xs_dict_append(acct, "id",           acct_md5);
     acct = xs_dict_append(acct, "username",     prefu);
     acct = xs_dict_append(acct, "display_name", display_name);
@@ -591,10 +602,48 @@ xs_dict *mastoapi_account(const xs_dict *actor)
         acct = xs_dict_append(acct, "emojis", eml);
     }
 
+    acct = xs_dict_append(acct, "locked", xs_stock_false);
+    acct = xs_dict_append(acct, "followers_count", xs_stock_0);
+    acct = xs_dict_append(acct, "following_count", xs_stock_0);
+    acct = xs_dict_append(acct, "statuses_count", xs_stock_0);
+
+    xs *fields = xs_list_new();
+    p = xs_dict_get(actor, "attachment");
+    xs_dict *v;
+
+    while (xs_list_iter(&p, &v)) {
+        char *type  = xs_dict_get(v, "type");
+        char *name  = xs_dict_get(v, "name");
+        char *value = xs_dict_get(v, "value");
+
+        if (!xs_is_null(type) && !xs_is_null(name) &&
+            !xs_is_null(value) && strcmp(type, "PropertyValue") == 0) {
+            xs *d = xs_dict_new();
+
+            d = xs_dict_append(d, "name", name);
+            d = xs_dict_append(d, "value", value);
+            d = xs_dict_append(d, "verified_at", xs_stock_null);
+
+            fields = xs_list_append(fields, d);
+        }
+    }
+
+    acct = xs_dict_append(acct, "fields", fields);
+
     return acct;
 }
 
 
+xs_str *mastoapi_date(char *date)
+/* converts an ISO 8601 date to whatever format Mastodon uses */
+{
+    xs_str *s = xs_crop_i(xs_dup(date), 0, 19);
+    s = xs_str_cat(s, ".000Z");
+
+    return s;
+}
+
+
 xs_dict *mastoapi_poll(snac *snac, const xs_dict *msg)
 /* creates a mastoapi Poll object */
 {
@@ -606,7 +655,8 @@ xs_dict *mastoapi_poll(snac *snac, const xs_dict *msg)
     xs *options = xs_list_new();
 
     poll = xs_dict_append(poll, "id", mid);
-    poll = xs_dict_append(poll, "expires_at", xs_dict_get(msg, "endTime"));
+    xs *fd = mastoapi_date(xs_dict_get(msg, "endTime"));
+    poll = xs_dict_append(poll, "expires_at", fd);
     poll = xs_dict_append(poll, "expired",
             xs_dict_get(msg, "closed") != NULL ? xs_stock_true : xs_stock_false);
 
@@ -660,7 +710,13 @@ xs_dict *mastoapi_status(snac *snac, const xs_dict *msg)
     const char *type = xs_dict_get(msg, "type");
     const char *id   = xs_dict_get(msg, "id");
 
+    /* fail if it's not a valid actor */
+    if (xs_is_null(type) || xs_is_null(id))
+        return NULL;
+
     xs *acct = mastoapi_account(actor);
+    if (acct == NULL)
+        return NULL;
 
     xs *idx = NULL;
     xs *ixc = NULL;
@@ -672,9 +728,11 @@ xs_dict *mastoapi_status(snac *snac, const xs_dict *msg)
     st = xs_dict_append(st, "id",           mid);
     st = xs_dict_append(st, "uri",          id);
     st = xs_dict_append(st, "url",          id);
-    st = xs_dict_append(st, "created_at",   xs_dict_get(msg, "published"));
     st = xs_dict_append(st, "account",      acct);
 
+    xs *fd = mastoapi_date(xs_dict_get(msg, "published"));
+    st = xs_dict_append(st, "created_at", fd);
+
     {
         const char *content = xs_dict_get(msg, "content");
         const char *name    = xs_dict_get(msg, "name");
@@ -927,10 +985,15 @@ xs_dict *mastoapi_status(snac *snac, const xs_dict *msg)
     st = xs_dict_append(st, "text", tmp);
 
     tmp = xs_dict_get(msg, "updated");
+    xs *fd2 = NULL;
     if (xs_is_null(tmp))
         tmp = xs_stock_null;
+    else {
+        fd2 = mastoapi_date(tmp);
+        tmp = fd2;
+    }
 
-    st = xs_dict_append(st, "edited_at", tmp);
+    st = xs_dict_append(st, "edited_at", fd2);
 
     if (strcmp(type, "Question") == 0) {
         xs *poll = mastoapi_poll(snac, msg);
@@ -1028,13 +1091,13 @@ int process_auth_token(snac *snac, const xs_dict *req)
                 /* this counts as a 'login' */
                 lastlog_write(snac, "mastoapi");
 
-                srv_debug(2, xs_fmt("mastoapi auth: valid token for user %s", uid));
+                srv_debug(2, xs_fmt("mastoapi auth: valid token for user '%s'", uid));
             }
             else
-                srv_log(xs_fmt("mastoapi auth: corrupted token %s", tokid));
+                srv_log(xs_fmt("mastoapi auth: corrupted token '%s'", tokid));
         }
         else
-            srv_log(xs_fmt("mastoapi auth: invalid token %s", tokid));
+            srv_log(xs_fmt("mastoapi auth: invalid token '%s'", tokid));
     }
 
     return logged_in;
@@ -1049,8 +1112,6 @@ int mastoapi_get_handler(const xs_dict *req, const char *q_path,
     if (!xs_startswith(q_path, "/api/v1/") && !xs_startswith(q_path, "/api/v2/"))
         return 0;
 
-    srv_debug(1, xs_fmt("mastoapi_get_handler %s", q_path));
-
     int status    = 404;
     xs_dict *args = xs_dict_get(req, "q_vars");
     xs *cmd       = xs_replace_n(q_path, "/api", "", 1);
@@ -1067,9 +1128,13 @@ int mastoapi_get_handler(const xs_dict *req, const char *q_path,
             acct = xs_dict_append(acct, "acct",         xs_dict_get(snac1.config, "uid"));
             acct = xs_dict_append(acct, "display_name", xs_dict_get(snac1.config, "name"));
             acct = xs_dict_append(acct, "created_at",   xs_dict_get(snac1.config, "published"));
+            acct = xs_dict_append(acct, "last_status_at", xs_dict_get(snac1.config, "published"));
             acct = xs_dict_append(acct, "note",         xs_dict_get(snac1.config, "bio"));
             acct = xs_dict_append(acct, "url",          snac1.actor);
             acct = xs_dict_append(acct, "header",       "");
+            acct = xs_dict_append(acct, "header_static", "");
+            acct = xs_dict_append(acct, "locked",       xs_stock_false);
+            acct = xs_dict_append(acct, "bot",          xs_dict_get(snac1.config, "bot"));
 
             xs *src = xs_json_loads("{\"privacy\":\"public\","
                     "\"sensitive\":false,\"fields\":[],\"note\":\"\"}");
@@ -1105,6 +1170,10 @@ int mastoapi_get_handler(const xs_dict *req, const char *q_path,
                 acct = xs_dict_set(acct, "fields", fields);
             }
 
+            acct = xs_dict_append(acct, "followers_count", xs_stock_0);
+            acct = xs_dict_append(acct, "following_count", xs_stock_0);
+            acct = xs_dict_append(acct, "statuses_count", xs_stock_0);
+
             *body  = xs_json_dumps(acct, 4);
             *ctype = "application/json";
             status = 200;
@@ -1143,6 +1212,33 @@ int mastoapi_get_handler(const xs_dict *req, const char *q_path,
             status = 422;
     }
     else
+    if (strcmp(cmd, "/v1/accounts/lookup") == 0) { /** **/
+        /* lookup an account */
+        char *acct = xs_dict_get(args, "acct");
+
+        if (!xs_is_null(acct)) {
+            xs *s = xs_strip_chars_i(xs_dup(acct), "@");
+            xs *l = xs_split_n(s, "@", 1);
+            char *uid = xs_list_get(l, 0);
+            char *host = xs_list_get(l, 1);
+
+            if (uid && (!host || strcmp(host, xs_dict_get(srv_config, "host")) == 0)) {
+                snac user;
+
+                if (user_open(&user, uid)) {
+                    xs *actor = msg_actor(&user);
+                    xs *macct = mastoapi_account(actor);
+
+                    *body  = xs_json_dumps(macct, 4);
+                    *ctype = "application/json";
+                    status = 200;
+
+                    user_free(&user);
+                }
+            }
+        }
+    }
+    else
     if (xs_startswith(cmd, "/v1/accounts/")) { /** **/
         /* account-related information */
         xs *l = xs_split(cmd, "/");
@@ -1342,12 +1438,16 @@ int mastoapi_get_handler(const xs_dict *req, const char *q_path,
                 if (!xs_match(type, "Note|Question|Page|Article"))
                     continue;
 
-                const char *from;
+                const char *from = NULL;
                 if (strcmp(type, "Page") == 0)
                     from = xs_dict_get(msg, "audience");
-                else
+
+                if (from == NULL)
                     from = xs_dict_get(msg, "attributedTo");
 
+                if (from == NULL)
+                    continue;
+
                 /* is this message from a person we don't follow? */
                 if (strcmp(from, snac1.actor) && !following_check(&snac1, from)) {
                     /* discard if it was not boosted */
@@ -1855,7 +1955,7 @@ int mastoapi_get_handler(const xs_dict *req, const char *q_path,
                     char *actor = NULL;
                     char *user  = NULL;
 
-                    if (valid_status(webfinger_request(q, &actor, &user))) {
+                    if (valid_status(webfinger_request(q, &actor, &user) && actor)) {
                         xs *actor_o = NULL;
 
                         if (valid_status(actor_request(&snac1, actor, &actor_o))) {
@@ -1883,6 +1983,8 @@ int mastoapi_get_handler(const xs_dict *req, const char *q_path,
     if (logged_in)
         user_free(&snac1);
 
+    srv_debug(1, xs_fmt("mastoapi_get_handler %s %d", q_path, status));
+
     return status;
 }
 
@@ -2391,6 +2493,27 @@ int mastoapi_post_handler(const xs_dict *req, const char *q_path,
 }
 
 
+int mastoapi_delete_handler(const xs_dict *req, const char *q_path,
+                             char **body, int *b_size, char **ctype) {
+
+    (void)req;
+    (void)body;
+    (void)b_size;
+    (void)ctype;
+
+    if (!xs_startswith(q_path, "/api/v1/") && !xs_startswith(q_path, "/api/v2/"))
+        return 0;
+
+    srv_debug(1, xs_fmt("mastoapi_delete_handler %s", q_path));
+    xs *cmd = xs_replace_n(q_path, "/api", "", 1);
+    if (xs_startswith(cmd, "/v1/push/subscription") || xs_startswith(cmd, "/v2/push/subscription")) { /** **/
+        // pretend we deleted it, since it doesn't exist anyway
+        return 200;
+    }
+    return 0;
+}
+
+
 int mastoapi_put_handler(const xs_dict *req, const char *q_path,
                           const char *payload, int p_size,
                           char **body, int *b_size, char **ctype)
